#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
#include "clobber.h"

#define EMPTY 0
#define OCCUPIED 1
#define RED 0
#define BLACK 1

list_type list_concat (list_type l1, list_type l2);
extern game_type clobber (void);
clobber_type clobber_make (int x, int y);
void clobber_free (clobber_type pos);
void clobber_copy_into (clobber_type b, clobber_type a);
clobber_type clobber_pack (clobber_type a);

extern game_type clobber (void) {
  clobber_type pos = clobber_input();
  return clobber_position (pos);
}

#define max(a, b)	((a) > (b) ? (a) : (b))
clobber_type clobber_input(void) {
    clobber_type pos;
    clobber_type newpos;

    char s[MAXBUFF];
    int i=0, j;
    
    pos = clobber_make (10,10);

    printf ("Enter `clobber' position, then hit <return> twice.\n");
    for (j=0; input_line(s) != NULL && *s != '\0';j++)
      for (i=0; *(s+i) != '\0'; i++) {
         if (i >= B_x (pos.occupied) || j >= B_y (pos.occupied)) {
           newpos = clobber_make (10+max(i,B_x(pos.occupied)),
                                  10+max(j, B_y(pos.occupied)));
           clobber_copy_into (newpos, pos);
           clobber_free (pos);
           pos = newpos;
         }
         if (*(s+i) == 'o' || *(s+i) == 'O') {
           B_set (pos.occupied, i, j);
         }
         if (*(s+i) == 'x' || *(s+i) == 'X') {
           B_set (pos.occupied, i, j);
           B_set (pos.colors, i, j);
         }
      }
    return pos;
}

void clobber_printf (clobber_type pos) {
  int x,y;
  for (y = 0; y < B_y (pos.occupied); y++) {
    for (x = 0; x < B_x (pos.occupied); x++)
      printf ("%s", B_get (pos.occupied, x, y) ?
              (B_get (pos.colors, x, y) ? "X" : "O") : " ");
    printf ("\n");
  }
}
  
int dir_x[]={-1,0,1,0};		/* the 4 directions */
int dir_y[]={0,-1,0,1};

game_type clobber_play (clobber_type pos, int x, int y, int adjx, int adjy) {
   clobber_type newpos;
   newpos.occupied = B_copy (pos.occupied);
   newpos.colors = B_copy (pos.colors);
   if (B_get (pos.colors, x, y))
      B_set (newpos.colors, adjx, adjy);
   else
      B_clear (newpos.colors, adjx, adjy);
   B_clear (newpos.occupied, x, y);
   return clobber_position (newpos);
}

game_type clobber_connected_position (clobber_type pos);
void clobber_copy_connected_component_at_ (clobber_type old,
                                           clobber_type New,
                                           int x, int y);
clobber_type clobber_copy_connected_component_at (clobber_type old,
                                            int x, int y);
extern game_type clobber_position (clobber_type pos) {
  game_type g=zero;
  int i, j;
  clobber_type New;	/* New is capitalized to avoid
			   the name conflict with C++ */
  for (j=0; j < B_y(pos.occupied); j++)
    for (i=0; i < B_x(pos.occupied); i++)
      if (B_get (pos.occupied, i, j)) {
        New = clobber_copy_connected_component_at (pos, i, j);
        g = plus (g, clobber_connected_position (New));
      }
  clobber_free (pos);
  return g;
}

game_type clobber_connected_position (clobber_type pos) {
   list_type left, right, lhash;
   int x = B_x (pos.occupied);
   int y = B_y (pos.occupied);
   int i, j, dir, adjx, adjy, col;
   game_type g;

   pos = clobber_pack (pos);
   lhash = list_concat (B_to_list (pos.occupied),
                        B_to_list (pos.colors));
   if (hash_test (CLOBBER_KEY, lhash))
     return clobber_free (pos), list_free (lhash), hash_get_last();
   
   // clobber_printf (pos);

   left = list_make();
   right = list_make();
   
   for (i = 0; i < x; i++)
      for (j = 0; j < y; j++)
         if (B_get (pos.occupied, i, j)) {
            col = B_get (pos.colors, i, j);
            for (dir = 0; dir < 4; dir++) {
               adjx = i+dir_x[dir];
               adjy = j+dir_y[dir];
               if (B_get (pos.occupied, adjx, adjy)
                   && (col != B_get (pos.colors, adjx, adjy)))
                  list_insert (col == BLACK ? left : right,
                               clobber_play (pos, i, j, adjx, adjy));
            }
         }
   g = make (left, right);
   hash_put (CLOBBER_KEY, lhash, g);
   clobber_free (pos);
   return g;
}

clobber_type clobber_copy_connected_component_at (clobber_type old,
                                            int x, int y)
{
    clobber_type New;
    
    New = clobber_make (B_x (old.occupied), B_y (old.occupied));
    clobber_copy_connected_component_at_ (old, New, x, y);
    return New;
}

void clobber_copy_connected_component_at_ (clobber_type old,
                                           clobber_type New,
                                           int x, int y)
{
    int i,j,n;

    B_clear (old.occupied, x, y);
    B_set (New.occupied, x, y);
    if (B_get (old.colors, x, y))
      B_set (New.colors, x, y);
    for (n = 0; n <= 3; n++) {
      i = x + dir_x[n];
      j = y + dir_y[n];
      if (B_get (old.occupied,i,j))
        clobber_copy_connected_component_at_ (old, New, i, j);
    }
}

clobber_type clobber_make (int x, int y) {
  clobber_type pos;
  pos.occupied = B_make (x, y);
  pos.colors = B_make (x, y);
  return pos;
}

void clobber_free (clobber_type pos) {
  B_free (pos.occupied);
  B_free (pos.colors);
}

void clobber_copy_into (clobber_type b, clobber_type a) {
  B_copy_into (b.occupied, a.occupied);
  B_copy_into (b.colors, a.colors);
}

clobber_type clobber_pack_copy (clobber_type a) {
  clobber_type b;
  int i, j, xmin, ymin, xmax, ymax;
  xmin=a.occupied->x, ymin=a.occupied->y, xmax=-1, ymax=-1;
  for (i=0; i < a.occupied->x; i++)
    for (j=0; j < a.occupied->y; j++)
      if (B_get (a.occupied,i,j)) {
        if (i < xmin) xmin = i;
        if (j < ymin) ymin = j;
        if (i > xmax) xmax = i;
        if (j > ymax) ymax = j;
      }
  if (xmin > xmax) {
    xmin = 0;
    ymin = 0;
  }
  b = clobber_make (xmax - xmin + 1, ymax - ymin + 1);
  for (i = 0; i < xmax - xmin + 1; i++)
    for (j = 0; j < ymax - ymin + 1; j++)
      if (B_get (a.occupied, i+xmin, j+ymin)) {
        B_set (b.occupied,i,j);
        if (B_get (a.colors, i+xmin, j+ymin))
          B_set (b.colors,i,j);
      }
  return b;
}

clobber_type clobber_pack (clobber_type a) {
  clobber_type b = clobber_pack_copy (a);
  clobber_free (a);
  return b;
}

list_type list_concat (list_type l1, list_type l2) {
  list_type tmp;
  list_type l;
  tmp = list_make();
  for (l = list_start (l1); l != NIL; l = list_rest (l))
    list_prepend (tmp, list_pick(l));
  list_free (l1);
  l1 = tmp;
  for (l = list_start (l1); l != NIL; l = list_rest (l))
    list_prepend (l2, list_pick(l));
  list_free (l1);
  return l2;
}

////////////////////////////////////////////////////////////////

#define max_clobber_games 1000

boolean next_clobber_position (clobber_type pos) {
  int i, j;
  for (j=0; j < B_y(pos.occupied); j++)
    for (i=0; i < B_x(pos.occupied); i++)
      if (B_get (pos.occupied, i, j)) {
	if (B_get (pos.colors, i, j)) {
	  B_clear (pos.colors, i, j);
	  B_clear (pos.occupied, i, j);
	} else {
	  B_set (pos.colors, i, j);
	  return TRUE;
	}
      } else {
	B_set (pos.occupied, i, j);
	B_clear (pos.colors, i, j);
	return TRUE;
      }
  return FALSE;
}

void clobber_search_for (int i, int j, game_type game) {
  clobber_type pos = clobber_make (i, j);
  game_type g, G, G_mean, G_norm;
  while(next_clobber_position (pos)) {
    clobber_type packed = clobber_pack_copy (pos);
    g = clobber_position (clobber_pack_copy (packed));
    G = atomic_weight (g);
    G_mean = num (mean (G));
    G_norm = minus (G, G_mean);
    if (eq (G_norm, game)) {
      printf ("-----------------\n");
      printf ("game = ");
      game_printf (g);
      printf ("atomic weight = ");
      game_print (G);
      printf (" = ");
      game_print (G_mean);
      printf (" + ");
      game_printf (G_norm);
      clobber_printf (packed);
      fflush (stdout);
    }
    clobber_free (packed);
  }
}

void clobber_search (int i, int j) {
  clobber_type pos = clobber_make (i, j);
  game_type g, G, G_mean, G_norm;
  int n = 0;
  int checked=0;
  game_type found_values[max_clobber_games];
  boolean star_found[100];
  for (i = 0; i < 100; i++)
    star_found[i] = FALSE;
  while(next_clobber_position (pos)) {
    clobber_type packed = clobber_pack_copy (pos);
    boolean found = FALSE;
    int i;

    checked++;
    if (checked % 1000 == 0) clean();
    g = clobber_position (clobber_pack_copy (packed));
    G = atomic_weight (g);
    G_mean = num (mean (G));
    G_norm = minus (G, G_mean);
    if (le (G_norm, zero)) found = TRUE;
    for (i = 0; i < n && !found; i++)
      if (eq (G_norm, found_values[i]))
	found = TRUE;
    if (!found) {
      found_values[n++] = G_norm;
      printf ("-----------------\n");
      printf ("game = ");
      game_printf (g);
      printf ("atomic weight = ");
      game_print (G);
      printf (" = ");
      game_print (G_mean);
      printf (" + ");
      game_printf (G_norm);
      clobber_printf (packed);
      fflush (stdout);
    }
    if (is_star (g) && !star_found[star_value(g)]) {
      star_found[star_value(g)] = TRUE;
      printf ("-----------------\n");
      printf ("game = ");
      game_printf (g);
      clobber_printf (packed);
    }
      
    clobber_free (packed);
  }
}

