#define JOIN_KEY 55
#define MEET_KEY 56
#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
#include "lattice.h"

#define DAYS 4
#define MAXLEVELS 50

game_type biggame;
list_type day[DAYS];
int today = 2;
list_type options[DAYS];
list_type *level;
int levels = 0;
int numlevels[DAYS] = {1, 3, 9, 45};
int numgames[DAYS] = {1, 4, 22, 1474};
#define subset_type game_type

// Call S a subset of incomparable games.  I'm encoding S
// in the game {bignum | S}, only because it's expedient.

list_type get_subset (subset_type g) {return right_options (g);}
void subset_printf (subset_type g) {games_printf (right_options (g));}
subset_type make_subset (list_type l) {
  return make (list_copy_1 (biggame), l);
}
subset_type union_subset (subset_type g, subset_type h) {
  return make_subset (list_union (get_subset (g), get_subset (h)));
}

void lattice_day (int day) {
  if (day < 0) printf ("Day must be non-negative\n");
  else if (day >= DAYS) printf ("Day cannot exceed %d\n", DAYS-1);
  else {
    today = day;
    printf ("Today is day %d\n", day);
  }
}
  
#undef max
int max (a, b) {return a > b ? a : b;}
#undef abs
int abs (a) {return a >= 0 ? a : -a;}
int birthday (game_type g) {
  list_type l;
  int day = -1;
  if (is_int (g)) return (abs (int_value (g)));
  for (l = list_start (left_options (g)); l != NIL; l = list_rest (l))
    day = max (day, birthday (list_pick (l)));
  for (l = list_start (right_options (g)); l != NIL; l = list_rest (l))
    day = max (day, birthday (list_pick (l)));
  return 1+day;
}

list_type antichains (list_type l) {
  // Returns a subset of 2^S, those subsets of S which contain only incomparable games.
  // Uses the following recurrence:
  // I({g} U S) = I(S) U {{g} U I(S\g)}, where S\g are games in S incomparable with g
  // I({}) = {{}}
   
  list_type l1, l2;
  game_type g;
  list_type result;

  l1 = list_copy (l);
  if (list_length (l1) == 0)
    return list_copy_1 (make_subset (list_make()));
  g = list_nth (l1, 1);
  list_delete_nth (l1, 1);
  l2 = list_make ();
  for (l = list_start (l1); l != NIL; l = list_rest (l)) {
    game_type h = list_pick(l);
    if (! (ge (g,h) || ge(h,g))) // g and h are incomparable
      list_insert (l2, h);
  }
  result = antichains (l1);
  l1 = antichains (l2);
  list_free (l2);
  for (l = list_start (l1); l != NIL; l = list_rest (l))
    list_insert (result, make_subset (list_union (list_1 (g),
                                                  get_subset (list_pick (l)))));
  list_free(l1);
  return result;
}

void lattice_init() {
  game_type g;
  list_type left, right;
  int n;
  int i,j,k;
  biggame = g_int(5);

  day[0] = list_copy_1(zero);
  for (i = 1; i < DAYS; i++) {
    options[i] = antichains (day[i-1]);
    // printf ("DAY %d has %d sets of options and ", i, list_length (options[i]));
    fflush (stdout);

    day[i] = list_make();
    
    n = list_length (options[i]);
    for (j = 1; j <= n; j++)
      for (k = n; k >= 1; k--) {
        left = list_copy (get_subset (list_nth (options[i], j)));
        right = list_copy (get_subset (list_nth (options[i], k)));
        g = make (left, right);
        if (!list_member (day[i], g))
          list_prepend (day[i], g);
      }
    // printf ("%d games\n", list_length (day[i]));
    fflush (stdout);
  }
}

void list_games() {
  int i, j;
  list_type l;
  for (i = 0; i < DAYS; i++) {
    printf ("DAY %d has %d games\n", i, list_length (day[i]));
    j = 1;
    for (l = list_start (day[i]); l != NIL; l = list_rest (l)) {
      printf ("g%d\t", j);
      game_printf (list_pick(l));
      j++;
    }
    printf ("\n");
  }
}

boolean B_equal (B_type b1, B_type b2, int n) {
  int i,j;
  for (i = 1; i <= n; i++)
    for (j = 1; j <= n; j++)
      if (B_get (b1, i, j) != B_get (b2, i, j))
        return FALSE;
  return TRUE;
}

void print_levels () {
  int i;
  list_type l;
  for (i = 1; i <= levels; i++) {
    printf ("Level %d: ", i);
    for (l = list_start (level[i]); l != NIL; l = list_rest (l)) {
      game_print (list_pick (l));
      printf (" ");
    }
    printf ("\n");
  }
}


void make_levels () {
  list_type list = list_copy (day[today]);
  list_type l;
  level = malloc (sizeof (list_type) * (numlevels[today] + 1));
  for (levels = 1; !list_empty (list); levels++) {
    level[levels] = list_copy (list);
    remove_dominated_options (level[levels], ge);
    for (l = list_start (level[levels]); l != NIL; l = list_rest (l))
      list_delete (list, list_pick(l));
  }
  levels--;
  list_free (list);
}  

B_type lattice_adjacencies () {
  B_type adj = B_make (numgames[today]+1, numgames[today]+1);
  list_type l1, l2;
  int n;
  int count = 0;
  for (n = 1; n < levels; n++)
    for (l1 = list_start (level[n]); l1 != NIL; l1 = list_rest (l1))
      for (l2 = list_start (level[n+1]); l2 != NIL; l2 = list_rest (l2))
        if (gt (list_pick (l1), list_pick (l2))) {
          B_set (adj,
                 list_search (day[today], list_pick (l1)),
                 list_search (day[today], list_pick (l2)));
          count++;
        }
  printf ("%d adjacencies\n", count);
  return adj;
}

B_type transitive_closure (B_type adj) {
  B_type result = B_copy (adj);
  boolean did_something = TRUE;
  int i, j, k;
  while (did_something) {
    did_something = FALSE;
    for (i = 0; i < B_x (result); i++) {
      printf (".");
      fflush (stdout);
      for (j = 0; j < B_x (result); j++)
        for (k = 0; k < B_x (result); k++)
          if (B_get (result, i, j)
              && B_get (result, j, k)
              && !B_get (result, i, k)) {
            B_set (result, i, k);
            did_something = TRUE;
          }
    }
  }
  return result;
}

boolean check_birthday (game_type g) {
  if (birthday (g) > today) {
    printf ("ERROR:  Birthday of ");
    game_print (g);
    printf (" exceeds ");
    printf ("%d\n", today);
    return FALSE;
  }
  return TRUE;
}

list_type lattice_floor (game_type g) {
  list_type l;
  list_type result = list_make();
  game_type h;
  for (l = list_start (day[today-1]); l != NIL; l = list_rest (l)) {
    h = list_pick(l);
    if (!ge (h, g))
      list_insert (result, h);
  }
  return result;
}

list_type lattice_ceil (game_type g) {
  list_type l;
  game_type h;
  list_type result = list_make();
  for (l = list_start (day[today-1]); l != NIL; l = list_rest (l)) {
    h = list_pick(l);
    if (!le (h, g))
      list_insert (result, h);
  }
  return result;
}

game_type join (game_type g1, game_type g2) {
  list_type left, right, l;
  game_type h;
  if (!(check_birthday (g1) && check_birthday (g2)))
    return NO_GAME;
  if (hash_test (JOIN_KEY, list_3 (g1, g2, today)))
    return hash_get_last();
  left = list_union (left_options (g1), left_options (g2));
  right = list_make();
  for (l = list_start (day[today-1]); l != NIL; l = list_rest (l)) {
    h = list_pick(l);
    if (!le (h, g1) && !le (h, g2))
      list_insert (right, h);
  }
  h = make (left, right);
  hash_put (JOIN_KEY, list_copy_3 (g1, g2, today), h);
  return h;
}

game_type meet (game_type g1, game_type g2) {
  list_type left, right, l;
  game_type h;
  if (!(check_birthday (g1) && check_birthday (g2)))
    return NO_GAME;
  if (hash_test (MEET_KEY, list_3 (g1, g2, today)))
    return hash_get_last();
  right = list_union (right_options (g1), right_options (g2));
  left = list_make();
  for (l = list_start (day[today-1]); l != NIL; l = list_rest (l)) {
    h = list_pick(l);
    if (!ge (h, g1) && !ge (h, g2))
      list_insert (left, h);
  }
  h = make (left, right);
  hash_put (MEET_KEY, list_copy_3 (g1, g2, today), h);
  return h;
}

void list_pseudocomplement (game_type a, game_type b) {
      game_print (a);
      printf (" -> ");
      game_print (b);
      printf (" = ");
      game_printf (pseudocomplement (a, b));
}  

void list_pseudocomplements_from (game_type g) {
  list_type l;
  if (!check_birthday (g)) return;
  for (l = list_start(day[today]); l != NIL; l = list_rest (l))
    list_pseudocomplement (g, list_pick (l));
  return;
}

void list_pseudocomplements_to (game_type g) {
  list_type l;
  if (!check_birthday (g)) return;
  for (l = list_start(day[today]); l != NIL; l = list_rest (l))
    list_pseudocomplement (list_pick (l), g);
  return;
}

void list_pseudocomplements () {
  list_type l1, l2;
  
  for (l1 = list_start(day[today]); l1 != NIL; l1 = list_rest (l1))
    for (l2 = list_start(day[today]); l2 != NIL; l2 = list_rest (l2))
      list_pseudocomplement (list_pick (l1), list_pick (l2));
}

game_type pseudocomplement (game_type a, game_type b) {
  // Returns the maximum x s.t. ax <= b
  list_type l;
  game_type x;
  game_type g = g_int (-today);
  if (!(check_birthday (a) && check_birthday (b)))
    return NO_GAME;
  for (l = list_start(day[today]); l != NIL; l = list_rest (l)) {
    x = list_pick (l);
    if (!le(x,g) && le(meet(a,x),b))
      g = join(x,g);
  }
  if (le (meet(a,g), b)) return g;
  else {
    printf ("Pseudocomplement undefined!\n");
    printf ("g = ");
    game_printf (g);
    return zero;
  }
}

list_type join_irreducible_elements () {
  boolean *element = malloc (sizeof (boolean) *
                             (numgames[today] + 1));
  list_type la, lb, result;
  int a, b, x;
  game_type gx, ga, gb;
  for (x = 1; x <= numgames[today]; x++)
    element[x] = TRUE;
  a = 0;
  for (a=1, la = list_start (day[today]); la != NIL; a++, la = list_rest (la))
    for (b=1, lb = list_start (day[today]); lb != NIL; b++, lb = list_rest (lb)) {
      ga = list_pick (la);
      gb = list_pick (lb);
      gx = join (ga, gb);
      if (! (eq (ga,gx) || eq(gb, gx)))
        element[list_search (day[today], gx)] = FALSE;
    }
  result = list_make();
  if (list_length (level[levels]) == 1)
    element[list_search (day[today], list_nth (level[levels], 1))] = FALSE;
  for (x = 1; x <= numgames[today]; x++)
    if (element[x]) list_insert (result, list_nth (day[today], x));
  return result;
}

void print_down_sets (list_type l) {
  list_type c, ll;
  int i;
  list_type ac = antichains (l);
  game_type g;
  printf ("%d antichains\n", list_length (ac));
  printf ("The corresponding downsets are:\n");
  for (c = list_start (ac); c != NIL; c = list_rest (c)) {
    printf ("{ ");
    fflush (stdout);
    for (i = 1; i <= list_length (l); i++) {
      g = list_nth (l, i);
      for (ll = list_start (get_subset (list_pick (c))); ll != NIL; ll = list_rest (ll))
        if (le (g, list_pick (ll))) {
          printf ("%d ", i);
          break;
        }
    }
    printf ("}\n");
    fflush (stdout);
  }
}

void lattice_test_1 () {
  list_type la, lb;
  list_type r;
  game_type ar, br;
  game_type a, b;

  for (la = list_start (day[today]); la != NIL; la = list_rest (la)) {
    a = list_pick(la);
    r = right_options (a);
    if (list_length (r) == 1) {
      ar = list_first (r);
      for (lb = list_start (day[today]); lb != NIL; lb = list_rest (lb)) {
        b = list_pick (lb);
        r = right_options (b);
        if (list_length (r) == 1) {
          br = list_first (r);
          if (list_member (lattice_ceil (br), ar)) {
            game_type x = pseudocomplement (a, b);
            if (!list_equal (right_options (x), right_options (b))) {
              printf ("Counterexample: ");
              game_print (a);
              printf (" -> ");
              game_print (b);
              printf (" = ");
              game_printf (x);
              return;
            }
          }
        }
      }
    }
  }
  printf ("Test passed");
}

void lattice_test (int n) {
  printf ("Running lattice test number %d...\n", n);
  if (n == 1) lattice_test_1();
}

/*
  int main() {
  list_type l;
  int i, j, k, len;
  boolean done;
  game_type gi, gj, gk;

  today = 3;
  init();
  lattice_init();
  make_levels ();

  printf ("%d levels\n", levels);

  j = 0;
  k = 0;
  for (i = 1; i <= numgames[today]; i++) {
  gi = list_nth (day[today], i);
  if (Q_eq (Q_int(0), temperature (gi))
  && (Q_eq (Q_int(0), mean (gi))))
  j++;
  if (is_all_small (gi)) k++;
  }
  printf ("%d infinitesimals are born on day %d\n", j, today);
  printf ("%d all-small games are born on day %d\n", k, today);

  l = join_irreducible_elements ();
  printf ("There are %d join-irreducible elements:\n", list_length (l));
   
  len = list_length (l);
  for (i=1; i <= len ; i++) {
  printf ("%d: ", i);
  game_printf (list_nth (l, i));
  }

  for (i = 1; i <= len; i++) {
  gi = list_nth (l, i);
  for (j = 1; j <= len; j++) {
  gj = list_nth (l, j);
  if (gt (gi, gj)) {
  done = FALSE;
  for (k = 1; k <= len && !done; k++) {
  gk = list_nth (l, k);
  if (gt (gi, gk) && gt (gk, gj)) done = TRUE;
  }
  if (!done) printf ("%d > %d\n", i, j);
  }
  }
  }

  // print_down_sets(l);
  return 0;
  }
*/
